/*
 * Segment 2 (KEx),  SWE2.4 Email/Fax/SMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.emailfaxsms;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;

import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;
import gnu.io.CommPort;
import gnu.io.CommPortIdentifier;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;
import gnu.io.UnsupportedCommOperationException;

/**
 * Realisiert die asynchrone Kommunikation mit der seriellen Schnittstelle.
 * Jeder Empfang oder Versand blockiert nur bis zu einem bestimmten Timeout, der
 * im Konstruktor angegeben wurde.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 */
public class AsyncSerialPortCommunicator implements SerialPortEventListener {

	/**
	 * Modemtyp.
	 */
	public enum ModemType {
		/** Festnetz-Modem. */
		MODEM_FESTNETZ, /** GSM-Modem. */
		MODEM_GSM, /** GPRS-Modem. */
		MODEM_GPRS,
	}

	/**
	 * Comm-Port Lock Objekt.
	 */
	private static HashMap<String, Object> comLock = new HashMap<String, Object>();

	/**
	 * Der Logger.
	 */
	protected static final Debug LOGGER = Debug.getLogger();

	/** OK-Antwort vom Modem. */
	public static final String OK = "OK";
	/** CONNECT-Meldung vom Modem. */
	public static final String CONNECT = "CONNECT";
	/** RING-Meldung vom Modem. */
	public static final String RING = "RING";
	/** NO_CARRIER-Meldung vom Modem. */
	public static final String NO_CARRIER = "NO<SP>CARRIER";
	/** ERROR-Meldung vom Modem. */
	public static final String ERROR = "ERROR";
	/** NO_DIALTONE-Meldung vom Modem. */
	public static final String NO_DIALTONE = "NO<SP>DIALTONE";
	/** BUSY-Meldung vom Modem. */
	public static final String BUSY = "BUSY";
	/** NO_ANSWER-Meldung vom Modem. */
	public static final String NO_ANSWER = "NO<SP>ANSWER";

	/** Wartezeit auf das Ende von Threads. */
	private static final int WARTEZEIT_THREAD_ENDE = 1000;

	/** Standard-Timeout. */
	private static final long DEFAULT_TIMEOUT = 30000L;

	/** Open-Timeout. */
	private static final int OPEN_TIMEOUT = 2000;

	/** Puffergr&ouml;&szlig;e. */
	private static final int PUFFER_GROESSE = 4096;

	/**
	 * Gibt das Lock-Objekt f&uuml;r den Port zur&uuml;ck.
	 *
	 * @param com
	 *            ComPort
	 * @return LockObjekt
	 */
	public static Object getLock(final String com) {
		synchronized (AsyncSerialPortCommunicator.comLock) {
			if (AsyncSerialPortCommunicator.comLock.containsKey(com)) {
				return AsyncSerialPortCommunicator.comLock.get(com);
			}
			AsyncSerialPortCommunicator.comLock.put(com, com);
			return com;
		}
	}

	/**
	 * Die Antworten vom Modem.
	 */
	private final String[] antworten = new String[] { AsyncSerialPortCommunicator.OK,
			AsyncSerialPortCommunicator.CONNECT, AsyncSerialPortCommunicator.RING,
			AsyncSerialPortCommunicator.NO_CARRIER, AsyncSerialPortCommunicator.ERROR,
			AsyncSerialPortCommunicator.NO_ANSWER, AsyncSerialPortCommunicator.NO_DIALTONE,
			AsyncSerialPortCommunicator.BUSY };

	/**
	 * Instanz der RX/TX seriellen Schnittstelle.
	 */
	protected SerialPort serialPort;

	/**
	 * Der Reader-Worker.
	 */
	protected AsyncSerialPortReader reader;

	/**
	 * Der Writer-Worker.
	 */
	protected AsyncSerialPortWriter writer;

	/**
	 * DEBUG.
	 */
	protected boolean debug = false;

	/**
	 * Defaultwert f&uuml;r timeout = 10 Sekunden.
	 */
	protected long timeout = AsyncSerialPortCommunicator.DEFAULT_TIMEOUT;

	/**
	 * Name des ComPorts.
	 */
	protected String com = new String("");

	/**
	 * Standardkonstruktor.
	 *
	 * @param com
	 *            Com-Port
	 * @param baud
	 *            Baudrate
	 * @param inPufferSize
	 *            Gr&ouml;sse des Input-Puffers
	 * @param outPufferSize
	 *            Gr&ouml;sse des Output-Puffers
	 * @param timeout
	 *            Timeout in Milisekunden
	 * @param swFlowControll
	 *            <code>true</code> wenn Software-Datenflusssteuerung benutzt
	 *            werden soll
	 * @throws Exception
	 *             Initialisierungsausnahme
	 */
	public AsyncSerialPortCommunicator(final String com, final int baud, final int inPufferSize,
			final int outPufferSize, final long timeout, final boolean swFlowControll) throws Exception {

		final CommPortIdentifier portIdentifier = CommPortIdentifier.getPortIdentifier(com);
		if (portIdentifier.isCurrentlyOwned()) {
			throw new PortBesetztException("Port " + com + " ist besetzt");
		}

		final CommPort commPort = portIdentifier.open("Kex Versand", AsyncSerialPortCommunicator.OPEN_TIMEOUT);
		if (!(commPort instanceof SerialPort)) {
			throw new Exception("Port " + com + " ist keine serielle Schnittstelle");
		}

		LogTools.log(AsyncSerialPortCommunicator.LOGGER, EmailFaxSmsMsg.FineComPortGeoeffnet, com);
		this.timeout = timeout;
		serialPort = (SerialPort) commPort;
		serialPort.setOutputBufferSize(AsyncSerialPortCommunicator.PUFFER_GROESSE);
		serialPort.setInputBufferSize(AsyncSerialPortCommunicator.PUFFER_GROESSE);
		serialPort.setDTR(true);

		serialPort.setSerialPortParams(baud, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);
		if (swFlowControll) {
			serialPort.setFlowControlMode(SerialPort.FLOWCONTROL_XONXOFF_OUT);
		} else {
			if (!serialPort.isCTS()) {
				serialPort.close();
				LogTools.log(AsyncSerialPortCommunicator.LOGGER, EmailFaxSmsMsg.FineComPortGeschlossen, com);
				throw new Exception("kein Gerät am Port: " + com + " (kein CTS)");
			}

			serialPort.setFlowControlMode(SerialPort.FLOWCONTROL_NONE);
			serialPort.setFlowControlMode(SerialPort.FLOWCONTROL_RTSCTS_OUT | SerialPort.FLOWCONTROL_RTSCTS_IN);
		}

		serialPort.addEventListener(this);
		serialPort.notifyOnDataAvailable(true);
		serialPort.notifyOnOutputEmpty(true);

		reader = new AsyncSerialPortReader(serialPort, inPufferSize);
		writer = new AsyncSerialPortWriter(serialPort, outPufferSize, reader, swFlowControll);

		reader.start();
		writer.start();

		// writer.setDebug(true);
		writer.swDatenFlussSteuerung = false;

		final long tstart = System.currentTimeMillis();
		do {
			if ((System.currentTimeMillis() - tstart) > 5000) {
				serialPort.close();
				throw new Exception("Thread-Start Fehler");

			}
			Thread.sleep(1000);

		} while (!reader.ready() || !writer.ready());

		LogTools.log(AsyncSerialPortCommunicator.LOGGER, EmailFaxSmsMsg.FineCommThreadsGestartet, com);

		this.com = com;
		synchronized (AsyncSerialPortCommunicator.comLock) {
			if (!AsyncSerialPortCommunicator.comLock.containsKey(com)) {
				AsyncSerialPortCommunicator.comLock.put(com, com);
			}
		}

	}

	/**
	 * Beendet die WorkerThreads und schliesst die Schnittstelle.
	 */
	public void beenden() {
		writer.beenden();
		reader.beenden();
		try {
			writer.join(AsyncSerialPortCommunicator.WARTEZEIT_THREAD_ENDE);
			reader.join(AsyncSerialPortCommunicator.WARTEZEIT_THREAD_ENDE);
		} catch (final InterruptedException e) {
			LogTools.log(AsyncSerialPortCommunicator.LOGGER, EmailFaxSmsMsg.WarningWartenThreadEndeUnterbrochen);
		}
		if (!writer.getState().equals(Thread.State.TERMINATED) || !reader.getState().equals(Thread.State.TERMINATED)) {
			LogTools.log(AsyncSerialPortCommunicator.LOGGER, EmailFaxSmsMsg.WarningWartenThreadEnde);
		}
		serialPort.close();
		LogTools.log(AsyncSerialPortCommunicator.LOGGER, EmailFaxSmsMsg.FineComPortGeschlossen, com);
	}

	/**
	 * Beendet den bin&auml;ren Modus. Soll bei Ende des bin&auml;ren Modus von
	 * der h&ouml;heren Schicht aufgerufen werden, da die Bedingung, wann der
	 * bin&auml;re Modus endet, vom Protokoll abh&auml;ngt (Fax Gr 2, Gr 2.0).
	 */
	public void endBinaerModus() {
		writer.endBinaryModus();
	}

	/**
	 * Gibtt den Namen des Comm-Ports zur&uuml;ck.
	 *
	 * @return Name des Comm-Ports
	 */
	public String getName() {
		return this.com;
	}

	/**
	 * Erfragt den eingestellten Timeout.
	 *
	 * @return Timeout
	 */
	public long getTimeout() {
		return timeout;
	}

	/**
	 * Wartet auf eine der erwarteten AT-Antworten oder Timeout.
	 *
	 * @return alle empfangenen Antworten
	 */
	public synchronized Collection<String> recvAlleATantworten() {
		final Collection<String> antwortenempfangen = new LinkedList<String>();
		for (;;) {
			final String s = reader.recvATResp(timeout);
			antwortenempfangen.add(s);
			for (final String t : antworten) {
				if (s == null) {
					return antwortenempfangen;
				} else if (t.equalsIgnoreCase(s)) {
					return antwortenempfangen;
				}
			}
		}
	}

	/**
	 * Wartet auf eine der erwarteten AT-Antworten oder Timeout.
	 *
	 * @return die Antwort
	 */
	public synchronized String recvATantwort() {
		for (;;) {
			final String s = reader.recvATResp(timeout);
			for (final String t : antworten) {
				if ((s == null) || t.equals(s)) {
					return s;
				}
			}
		}
	}

	/**
	 * Wartet auf eine der erwarteten AT-Antworten oder einen bestimmten
	 * Timeout.
	 *
	 * @param waittimeout
	 *            Timeout
	 * @return die Antwort
	 */
	public synchronized String recvATantwort(final long waittimeout) {
		for (;;) {
			final String s = reader.recvATResp(waittimeout);
			for (final String t : antworten) {
				if ((s == null) || t.equals(s)) {
					return s;
				}
			}
		}
	}

	/**
	 * Wartet bis max Timeout gekommen ist, oder eine von der erkannten AT
	 * Antworten, Im Fall Timeout gekommen ist oder die Antwort ist negativ,
	 * dann wird ein Exception geworfen sonst alle antworte die inzwischen
	 * gekommen sind auser der ersten, die das Echo von dem Kommando ist.
	 *
	 * @return alle Antworte die gekommen sind
	 * @throws CommunicationException
	 *             bei Ausnahmen
	 */
	public synchronized Collection<String> recvATantwortMitKontroll() throws CommunicationException {
		final Collection<String> antworte = new LinkedList<String>();
		for (;;) {
			final String s = reader.recvATResp(timeout);
			antworte.add(s);
			for (final String t : antworten) {
				if (s == null) {
					throw new CommunicationException("Timeout");
				} else if (t.equals(s)) {
					if (t.equals(AsyncSerialPortCommunicator.OK) || t.equals(AsyncSerialPortCommunicator.CONNECT)) {
						return antworte;
					}
					throw new CommunicationException(t);
				}
			}
		}
	}

	/**
	 * Empf&auml;ngt eine AT-Antwort.
	 *
	 * @return Antwort
	 */
	public String recvATResp() {
		return reader.recvATResp(timeout);
	}

	/**
	 * Wartet bis die erwartete ByteKette gekommen ist, oder timeout.
	 *
	 * @param erwartet
	 *            erwartete ByteKette
	 * @return true, wenn bytette gekommen ist
	 */
	public boolean recvBytes(final byte[] erwartet) {
		return reader.recvBytes(timeout, erwartet);
	}

	/**
	 * Versendet ein Byte ohne modification.
	 *
	 * @param b
	 *            ByteKommando
	 * @return ob der Versand erfolgreich war
	 */
	public boolean sendByte(final byte b) {
		return writer.sendByte(timeout, b);
	}

	/**
	 * Versendet Bytes ohne Modifikation.
	 *
	 * @param cmd
	 *            BytesKommando
	 * @return ob der Versand erfolgreich war
	 */
	public boolean sendBytes(final byte[] cmd) {
		final boolean x = writer.sendBytes(timeout, cmd, 0, cmd.length);
		return x;
	}

	/**
	 * Versendet Bytes ohne Modifikation.
	 *
	 * @param cmd
	 *            BytesKommando
	 * @param off
	 *            Offset
	 * @param len
	 *            Length
	 * @return ob der Versand erfolgreich war
	 */
	public boolean sendBytes(final byte[] cmd, final int off, final int len) {
		final boolean x = writer.sendBytes(timeout, cmd, off, len);
		return x;
	}

	/**
	 * Versendet ein AT-Kommando.
	 *
	 * @param waittimeout
	 *            Timeout
	 * @param cmd
	 *            AT Kommando
	 * @return ob der Versand erfolgreich war
	 */
	public boolean sendCmd(final long waittimeout, final String cmd) {
		return writer.sendCmd(waittimeout, cmd);
	}

	/**
	 * Versendet ein AT-Kommando.
	 *
	 * @param cmd
	 *            AT Kommando
	 * @return ob der Versand erfolgreich war
	 */
	public boolean sendCmd(final String cmd) {
		return writer.sendCmd(timeout, cmd);
	}

	/**
	 * Ein Wrapper f&uuml;r sendCmd + recvATantwortMitKontroll().
	 *
	 * @param cmd
	 *            Kommando
	 * @return Antwort
	 * @throws CommunicationException
	 *             Fehler
	 */
	public Collection<String> sendCmdMitKontroll(final String cmd) throws CommunicationException {
		Collection<String> resp = null;
		sendCmd(cmd);
		try {
			resp = recvATantwortMitKontroll();
		} catch (final CommunicationException e) {
			throw new CommunicationException(cmd, e.getMessage());
		}
		return resp;
	}

	/**
	 * Versendet Daten wobei die DLEs verdoppelt werden.
	 *
	 * @param data
	 *            Daten
	 * @param off
	 *            Offset
	 * @param len
	 *            L&auml;nge
	 * @param reverseBitOrder
	 *            Flag, ob die Bitorder umgekehrt werden soll
	 * @return in Anzahl der versendeten Bytes
	 */
	public int sendData(final byte[] data, final int off, final int len, final boolean reverseBitOrder) {
		try {
			return writer.sendData(timeout, data, off, len, reverseBitOrder);
		} catch (final Exception e) {
			e.printStackTrace();
		}
		return 0;
	}

	@Override
	public void serialEvent(final SerialPortEvent event) {

		switch (event.getEventType()) {
		case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
			writer.outputPufferLeer();
			break;
		case SerialPortEvent.DATA_AVAILABLE:
			if (!reader.getStoppend()) {
				reader.datenEingekommen();
			}
			break;
		default:
			break;
		}
	}

	/**
	 * Setzt den Debug Modus.
	 *
	 * @param debug
	 *            true/false
	 */
	public void setDebug(final boolean debug) {
		this.debug = debug;
		writer.setDebug(debug);
		reader.setDebug(debug);
	}

	/**
	 * Stellt den Timeout ein.
	 *
	 * @param t
	 *            Timeout in ms
	 */
	public void setTimeout(final long t) {
		timeout = t;
	}

	/**
	 * Startet den bin&auml;ren Modus. Soll bei Beginn des bin&auml;ren Modus
	 * von der h&ouml;heren Schicht aufgerufen werden, da die Bedingung, wann
	 * der bin&auml;re Modus beginnt, vom Protokoll abh&auml;ngt (Fax Gr 2, Gr
	 * 2.0).
	 */
	public void startBinaerModus() {
		// writer.startBinaryModus();
		final byte[] erw = new byte[] { AsyncSerialPort.XON };
		if (!reader.recvBytes(5000, erw)) {
			System.out.println("kein XON");
		}

		try {
			serialPort.setFlowControlMode(SerialPort.FLOWCONTROL_XONXOFF_OUT | SerialPort.FLOWCONTROL_XONXOFF_IN
					| SerialPort.FLOWCONTROL_RTSCTS_IN | SerialPort.FLOWCONTROL_RTSCTS_OUT);
		} catch (final UnsupportedCommOperationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Testausgabe Schnittstellenparameter.
	 */
	public void test() {

		final int x = serialPort.getFlowControlMode();

		System.err.println("Flow control " + x + "\nNONE: " + (SerialPort.FLOWCONTROL_NONE & x) + "\nRTS_CTS_IN: "
				+ (SerialPort.FLOWCONTROL_RTSCTS_IN & x) + "\nRTS_CTS_Out: " + (SerialPort.FLOWCONTROL_RTSCTS_OUT & x)
				+ "\nXON_XOFF_IN: " + (SerialPort.FLOWCONTROL_XONXOFF_IN & x) + "\nXON_XOFF_OUT: "
				+ (SerialPort.FLOWCONTROL_XONXOFF_OUT & x));

	}
}
