/*
 * Segment 2 (KEx),  SWE2.4 Email/Fax/SMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.emailfaxsms.fax.sender;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collection;

import de.bsvrz.kex.emailfaxsms.AsyncSerialPort;
import de.bsvrz.kex.emailfaxsms.AsyncSerialPortCommunicator;
import de.bsvrz.kex.emailfaxsms.AsyncSerialPortCommunicator.ModemType;
import de.bsvrz.kex.emailfaxsms.CommunicationException;
import de.bsvrz.kex.emailfaxsms.EmailFaxSmsMsg;
import de.bsvrz.kex.emailfaxsms.VersandStatus;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import gnu.io.NoSuchPortException;

/**
 * Die FaxVersand Klasse, Versendet eine Mitteilung.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 */
public class KexFaxSender {

	/**
	 * Fax Klasse.
	 */
	public enum FaxClass {
		/**
		 * Fax-Klasse 2.
		 */
		FAX_CLASS_2, /**
		 * Fax-Klasse 2.0.
		 */
		FAX_CLASS_2_0,
	}

	/**
	 * LOGGER.
	 */
	protected static final Debug LOGGER = Debug.getLogger();
	/**
	 * Konstanten.
	 */
	/** KByte. */
	protected static final int KILO_BYTE = 1024;
	/** Carriage-return. */
	protected static final byte CR = 0x0D;
	/** Line-feed. */
	protected static final byte LF = 0x0A;
	/** Endline. */
	protected static final byte[] ENDL = new byte[] { KexFaxSender.CR, KexFaxSender.LF };

	// Data-Link-Escape
	// protected final byte DLE = 0x10;
	/** End-of-Trasnmission. */
	protected static final byte ETX = 0x03;
	/** End-of-page. */
	protected static final byte EOP = 0x2e;
	/** End-of-message. */
	protected static final byte EOM = 0x3b;

	/** More-Pages. */
	protected static final byte MPS = 0x2c;
	/**
	 * Headersize in G3 Fax Files generiert mit GhostScript ueber dem
	 * dfaxhigh/dfaxlow device.
	 */
	protected static final int HDR_SIZE = 64;
	/**
	 * Antwort auf +CPIN?, wenn SIM Karte Aktiviert ist.
	 */
	protected static final String READY = "+CPIN:<SP>READY";
	/**
	 * Antwort auf +CPIN?, wenn SIM Karte gesperrt ist.
	 */
	protected static final String SIM_PIN = "+CPIN:<SP>SIM<SP>PIN";
	/**
	 * Puffergroesse.
	 */
	protected static final int PUFFER_SIZE = 4 * KexFaxSender.KILO_BYTE;

	/**
	 * Konstanten aus dem Multitech Fax developers Manual
	 * http://www.multitech.com/DOCUMENTS/Collateral/manuals/S000239C.pdf und
	 * Cisco
	 * http://www.cisco.com/univercd/cc/td/doc/product/access/acs_mod/cis2600/
	 * sw_conf/firmware/atwicbk/atwic5.htmhttp://www.multitech.com/DOCUMENTS/
	 * Collateral/manuals/S000239C.pdf
	 */
	/** Fax 2 EOP. */
	protected static final byte[] FAX_2_EOP = new byte[] { AsyncSerialPort.DLE, KexFaxSender.ETX };
	/** Letzte Seite. */
	protected static final byte[] FAX_20_EOP = new byte[] { AsyncSerialPort.DLE, KexFaxSender.EOP };
	/** Seite Ende, es kommt noch eine andere. */
	protected static final byte[] FAX_20_MPS = new byte[] { AsyncSerialPort.DLE, KexFaxSender.MPS };

	/**
	 * Standardtimeout.
	 */
	protected long timeout = 20000;

	/**
	 * Pin-Kode fuer die Authentifizierung im GSM-Netz.
	 */
	protected String gsmPin;
	/**
	 * Initialisierungsequenz.
	 */
	protected String initSeq;
	/**
	 * Prefix fuer alle Nummer.
	 */
	protected String prefix;

	/**
	 * Typ des Faxgeraetes.
	 */
	protected ModemType modemType = ModemType.MODEM_FESTNETZ;
	/**
	 * ComPort.
	 */
	protected String com;
	/**
	 * Baudrate.
	 */
	protected int baudrate;

	/**
	 * Puffer.
	 */
	protected byte[] buffer = new byte[KexFaxSender.PUFFER_SIZE];

	/**
	 * Standardkonstruktor.
	 *
	 * @param com
	 *            Com-Port
	 * @param baud
	 *            Baudrate
	 * @param timeout
	 *            Timeout in Sekunden
	 * @param typ
	 *            Modem Typ
	 * @param gsmPin
	 *            Pin Code
	 * @param init
	 *            Init Sequenz
	 * @param prefix
	 *            Wahl-Praefix
	 * @throws Exception
	 *             Wenn kein Zugriff zum COM-Port moeglich ist
	 */
	public KexFaxSender(final String com, final int baud, final long timeout, final ModemType typ, final String gsmPin,
			final String init, final String prefix) throws Exception {
		this.timeout = timeout * 1000;
		this.modemType = typ;
		this.gsmPin = gsmPin;
		this.initSeq = init;
		this.prefix = prefix;
		this.com = com;
		this.baudrate = baud;

		if ((this.initSeq == null) || (this.initSeq.length() == 0)) {
			this.initSeq = "ATZ";
		}

		if ((this.prefix == null) || (this.prefix.length() == 0)) {
			this.prefix = "ATD";
		}

		AsyncSerialPortCommunicator comm;
		// Nur Test Phase - kleiner Timeout
		try {
		comm = new AsyncSerialPortCommunicator(com, baud, 10 * KexFaxSender.KILO_BYTE, 10 * KexFaxSender.KILO_BYTE,
				5000, false);
		} catch (final UnsatisfiedLinkError ule) {
			KexFaxSender.LOGGER.error(ule.getMessage());
			throw new NoSuchPortException();
		}
		
		try {
			comm.sendCmdMitKontroll("AT");
		} catch (final CommunicationException e) {
			KexFaxSender.LOGGER.warning("Keine Antwort des Kommunikationsgerätes an " + com);
			comm.beenden();
			throw e;
		}

		comm.beenden();
		KexFaxSender.LOGGER.fine("FaxSender Initialisierung komplett");
	}

	// /**
	// * @param buffer1
	// * @param buffer2
	// * @param n
	// * @return
	// */
	// public int fixBuffer(final byte[] buffer1, final byte[] buffer2, final
	// int n) {
	// int i, j = 0;
	// for (i = 0; i < n; i++, j++) {
	// buffer2[j] = buffer1[i];
	// if (buffer2[j] == AsyncSerialPort.DLE) {
	// buffer2[++j] = AsyncSerialPort.DLE;
	// }
	// }
	// buffer2[j++] = AsyncSerialPort.DLE;
	// buffer2[j++] = ETX;
	// return j;
	// }

	/**
	 * Versendet eine Mitteilung.
	 *
	 * @param faxKennung
	 *            FaxKennung des Versenders
	 * @param nummer
	 *            Nummer des Empfaengers (ohne Standardpraefix)
	 * @param hochAufloesend
	 *            true, wenn mitteilung Hochaufloesend ist
	 * @param rasterGrafik
	 *            File, das die Mitteilung enthaelt
	 * @return VersandStatus
	 */
	public VersandStatus send(final String faxKennung, final String nummer, final boolean hochAufloesend,
			final File rasterGrafik) {
		VersandStatus status = VersandStatus.NICHT_VERSENDET;
		AsyncSerialPortCommunicator comm = null;
		synchronized (AsyncSerialPortCommunicator.getLock(com)) {
			try {
				KexFaxSender.LOGGER.fine("Beginne Fax-Versand an: " + nummer);

				comm = new AsyncSerialPortCommunicator(com, baudrate, 10 * KexFaxSender.KILO_BYTE,
						10 * KexFaxSender.KILO_BYTE, this.timeout, false);

				FaxClass faxClass = FaxClass.FAX_CLASS_2_0;
				// Typ Detektion (wahrscheinlich sind alle schon 2.0)
				try {
					comm.sendCmdMitKontroll("AT");
					comm.sendCmdMitKontroll("AT+FCLASS=2.0");
				} catch (final CommunicationException e) {
					faxClass = FaxClass.FAX_CLASS_2;
				}

				if (faxClass == FaxClass.FAX_CLASS_2_0) {
					status = send20(comm, faxKennung, prefix + nummer, hochAufloesend, rasterGrafik);
				} else {
					status = send2(comm, faxKennung, prefix + nummer, hochAufloesend, rasterGrafik);
				}

				comm.beenden();

			} catch (final Exception e) {
				final String s = "FAX nicht versendet: " + e.getMessage();
				KexFaxSender.LOGGER.warning(s);
				MessageSender.getInstance().sendMessage(MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, s);
				if (comm != null) {
					comm.beenden();
				}
				return new VersandStatus(s, false);
			}
		}
		return status;
	}

	/**
	 * Versendet eine Mitteilung mit dem Protokol der Fax Klasse 2.
	 *
	 * @param comm
	 *            ComPort
	 * @param faxKennung
	 *            FaxKennung des Versenders
	 * @param waehlstring
	 *            W&auml;hlstring (Prefix+Nummer)
	 * @param hochAufloesend
	 *            true, wenn mitteilung Hochaufloesend ist
	 * @param rasterGrafik
	 *            File, das die Mitteilung enth&auml;lt
	 * @return VersandStatus
	 */
	public VersandStatus send2(final AsyncSerialPortCommunicator comm, final String faxKennung,
			final String waehlstring, final boolean hochAufloesend, final File rasterGrafik) {

		int n, m;
		BufferedInputStream in;

		KexFaxSender.LOGGER.finer("Sende als FAX-CLASS2");

		try {
			in = new BufferedInputStream(new FileInputStream(rasterGrafik));
		} catch (final FileNotFoundException e) {
			KexFaxSender.LOGGER
					.warning("Fehler: Rastergrafik der Faxmitteilung konnte nicht gelesen werden: " + e.getMessage());
			return VersandStatus.NICHT_VERSENDET;
		}
		boolean versendet = false;
		Collection<String> resp;
		try {
			comm.sendCmdMitKontroll(initSeq);

			setCpinReady(comm, this.gsmPin);

			if (modemType == ModemType.MODEM_GSM) {
				// GSM-Modem (Siemens) kann ATS2 nicht)
				comm.sendCmdMitKontroll("ATE0Q0V1S0=0");
			} else {
				comm.sendCmdMitKontroll("ATE0Q0V1S0=0S2=43");
			}

			comm.sendCmdMitKontroll("AT+FCLASS=2");
			comm.sendCmdMitKontroll("AT+FLID=\"" + faxKennung + "\"");
			// TEST
			// comm.sendCmdMitKontroll("AT+FDIS=" + (hochAufloesend ? "1" : "0")
			// + ",5,0,2,0,0,0,0");
			comm.sendCmdMitKontroll("AT+FDCC=" + (hochAufloesend ? "1" : "0") + ",3,0,2,0,0,0,0"); // SIEMENS
			// MC35
			// + ",5,0,2,0,0,0,0");

			// mit den getesteten Modem funktioniert das nur mit FBOR=0 und
			// reverseBitOrder
			// comm.sendCmdMitKontroll("AT+FBOR=1");
			comm.sendCmdMitKontroll("AT+FBOR=0");
			comm.sendCmdMitKontroll(waehlstring);
			comm.sendCmdMitKontroll("AT+FDT");

			comm.startBinaerModus();
			in.skip(KexFaxSender.HDR_SIZE);
			LogTools.log(KexFaxSender.LOGGER, EmailFaxSmsMsg.FineSendeBinaerBytes, in.available(), rasterGrafik);
			while (in.available() > 0) {
				n = in.read(buffer);
				m = comm.sendData(buffer, 0, n, true);
				if (m != n) {
					throw new CommunicationException(
							"Es konnten nicht alle Daten versendet werden: AUFTRAG: " + n + " VERSENDET: " + m);
				}
			}
			in.close();
			comm.sendBytes(KexFaxSender.FAX_2_EOP);
			comm.endBinaerModus();

			comm.recvATantwortMitKontroll();
			resp = comm.sendCmdMitKontroll("AT+FET=2");
			versendet = ueberPruefeAntwort(resp, "+FPTS:", "1", "Daten von der Gegenstelle nicht korrekt empfangen");
			versendet = ueberPruefeAntwort(resp, "+FHNG:", "0", "Fehler in Phase D");

			// neu: auf 'no carrier warteb
			comm.recvATantwort(5000);

			comm.sendCmdMitKontroll("ATZ");
		} catch (final IOException e) {
			final String s = "Fehler beim Lesen der Rastergrafik: " + e.getMessage();
			KexFaxSender.LOGGER.warning(s);
			MessageSender.getInstance().sendMessage(MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, s);
			comm.sendCmd(1000, "AT+FAXERR?");
			comm.recvATantwort(1000);
			comm.sendCmd(1000, "ATH0");
			comm.recvATantwort(1000);
			return VersandStatus.NICHT_VERSENDET;
		} catch (final CommunicationException e) {
			final String s = "Fax nicht versendet: " + e.getMessage();
			KexFaxSender.LOGGER.warning(s);
			comm.sendCmd(1000, "AT+FAXERR?");
			comm.recvATantwort(1000);
			comm.sendCmd(1000, "ATH0");
			comm.recvATantwort(1000);
			MessageSender.getInstance().sendMessage(MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, s);
			if (versendet) {
				return new VersandStatus(
						"Fax von Gegenstelle korrekt empfangen, obwohl Fehler eingetreten: " + e.getMessage(), true);
			}
			return new VersandStatus("Kommunikationsfehler: " + e.getMessage(), false);
		}
		return VersandStatus.VERSAND_OK;
	}

	/**
	 * Versendet eine Mitteilung mit dem Protokol der Fax Klasse 2.0.
	 *
	 * @param comm
	 *            ComPort
	 * @param faxKennung
	 *            FaxKennung des Versenders
	 * @param waehlstring
	 *            W&auml;hlstring (Prefix+Nummer)
	 * @param hochAufloesend
	 *            true, wenn mitteilung Hochaufloesend ist
	 * @param rasterGrafik
	 *            File, das die Mitteilung enthaelt
	 * @return VersandStatus
	 */
	public VersandStatus send20(final AsyncSerialPortCommunicator comm, final String faxKennung,
			final String waehlstring, final boolean hochAufloesend, final File rasterGrafik) {

		int n, m;
		BufferedInputStream in;

		KexFaxSender.LOGGER.finer("Sende als FAX-CLASS2.0");

		try {
			in = new BufferedInputStream(new FileInputStream(rasterGrafik));
		} catch (final FileNotFoundException e) {
			KexFaxSender.LOGGER
					.warning("Fehler: Rastergrafik der Faxmitteilung konnte nicht gelesen werden: " + e.getMessage());
			return VersandStatus.NICHT_VERSENDET;
		}

		Collection<String> resp;
		boolean versendet = false;
		try {

			comm.sendCmdMitKontroll(initSeq);

			setCpinReady(comm, this.gsmPin);

			if (modemType == ModemType.MODEM_GSM) {
				// GSM-Modem (Siemens) kann ATS2 nicht)
				comm.sendCmdMitKontroll("ATE0Q0V1S0=0");
			} else {
				comm.sendCmdMitKontroll("ATE0Q0V1S0=0S2=43");
			}

			comm.sendCmdMitKontroll("AT+FCLASS=2.0");
			comm.sendCmdMitKontroll("AT+FCC=" + (hochAufloesend ? "1" : "0") + ",5,0,2,0,0,0,0");
			comm.sendCmdMitKontroll("AT+FBO=1");
			comm.sendCmdMitKontroll("AT+FNR=1,1,1,1");
			comm.sendCmdMitKontroll("AT+FLI=\"" + faxKennung + "\"");
			comm.sendCmdMitKontroll(waehlstring);
			comm.sendCmdMitKontroll("AT+FDT");

			comm.startBinaerModus();
			in.skip(KexFaxSender.HDR_SIZE);

			LogTools.log(KexFaxSender.LOGGER, EmailFaxSmsMsg.FineSendeBinaerBytes, in.available(), rasterGrafik);

			while (in.available() > 0) {
				n = in.read(buffer);
				m = comm.sendData(buffer, 0, n, false);
				if (m != n) {
					throw new CommunicationException(
							"Es konnten nicht alle Daten versendet werden: AUFTRAG: " + n + " VERSENDET: " + m);
				}
			}

			comm.sendBytes(KexFaxSender.FAX_20_EOP);
			// comm.sendData(FAX_20_EOP, 0, FAX_20_EOP.length);
			comm.endBinaerModus();

			resp = comm.recvAlleATantworten();

			// das kommt beim getesteten ISDN-Modem nicht
			// versendet = ueberPruefeAntwort(resp, "+FPS:<SP>1",
			// "Daten vom Remote nicht korrekt empfangen");
			versendet = ueberPruefeAntwort(resp, "+FHS:00", "Fehler in Phase D");
			//
			// versendet = ueberPruefeAntwort(resp, "+FHS:<SP>0",
			// "Fehler in Phase D");

			comm.sendCmdMitKontroll("ATH0");
			in.close();

		} catch (final IOException e) {
			final String s = "Fehler beim Lesen der Rastergrafik: " + e.getMessage();
			KexFaxSender.LOGGER.warning(s);
			MessageSender.getInstance().sendMessage(MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, s);
			comm.sendCmd(1000, "AT+FHS?");
			comm.recvATantwort(1000);
			comm.sendCmd(1000, "ATH0");
			comm.recvATantwort(1000);
			comm.recvATantwort();
			return VersandStatus.NICHT_VERSENDET;
		} catch (final CommunicationException e) {
			comm.sendCmd(1000, "AT+FHS?");
			comm.recvATantwort(1000);
			comm.sendCmd(1000, "ATH0");
			comm.recvATantwort(1000);
			final String s = "Fax nicht versendet: " + e.getMessage();
			KexFaxSender.LOGGER.warning(s);
			MessageSender.getInstance().sendMessage(MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, s);
			if (versendet) {
				return new VersandStatus(
						"Fax von Gegenstelle korrekt empfangen, obwohl Fehler aufgetreten: " + e.getMessage(), true);
			}
			return new VersandStatus("Kommunikationsfehler: " + e.getMessage(), false);
		}
		return VersandStatus.VERSAND_OK;
	}

	/**
	 * Entsperrt ein GSM-Faxmodem.
	 *
	 * @param comm
	 *            ComPort
	 * @param pin
	 *            Pin-Code
	 * @throws CommunicationException
	 *             bei Ausnahmen
	 */
	public void setCpinReady(final AsyncSerialPortCommunicator comm, final String pin) throws CommunicationException {
		if (modemType == ModemType.MODEM_FESTNETZ) {
			return;
		}

		final String cPIN = "+CPIN";
		boolean cpinUnterstuetzt = false;
		Collection<String> antwort = null;
		try {
			comm.sendCmd("AT+CPIN?");
			antwort = comm.recvATantwortMitKontroll();
		} catch (final CommunicationException e) {
			// Modem unterstuetzt nicht CPIN d.H. es ist kein GSM modem
			if (modemType == ModemType.MODEM_GSM) {
				KexFaxSender.LOGGER.warning("Modemtyp ist GSM, aber das Geraet unterstützt das Kommando +CPIN nicht");
			}
			return;
		}

		for (final String s : antwort) {
			if (s.startsWith(KexFaxSender.READY)) {
				return;
			} else if (s.startsWith(KexFaxSender.SIM_PIN)) {
				try {
					comm.sendCmd("AT+CPIN=" + pin);
					comm.recvATantwortMitKontroll();
					// Nach der entsperrung, wenn man sofort eine SMS versendet
					// bekommt man oft ein ERRORs
					Thread.sleep(10000);
				} catch (final CommunicationException e) {
					throw new CommunicationException("SIM-Karte konnte nicht entsperrt werden");
				} catch (final InterruptedException e) {
					KexFaxSender.LOGGER.warning("Sleep nach PIN-Eingabe wurde unterbrochen");
				}
				return;
			} else if (s.startsWith(cPIN)) {
				cpinUnterstuetzt = true;
			}
		}
		if (!cpinUnterstuetzt) {
			return;
		}

		throw new CommunicationException("SIM Karte konnte nicht entsperrt werden");
	}

	/**
	 * &Uuml;berpr&uuml;ft ob die Antwort einen gesuchten String enth&auml;lt.
	 *
	 * @param antwort
	 *            Antwort des Faxgeraets
	 * @param gesucht
	 *            Gesuchte Zeichenkette
	 * @param fehlermeldung
	 *            Fehlermeldung, wenn gesuchte Zeichenkette nicht gefunden ist
	 * @return true, wenn gesuchte Zeichenkette gefunden ist
	 * @throws CommunicationException
	 *             wird geworfen, wenn gesuchte Zeichenkette nicht gefunden ist
	 */
	public boolean ueberPruefeAntwort(final Collection<String> antwort, final String gesucht,
			final String fehlermeldung) throws CommunicationException {
		for (final String s : antwort) {
			if (s.startsWith(gesucht)) {
				return true;
			}
		}
		throw new CommunicationException(fehlermeldung);
	}

	/**
	 * &Uuml;berpr&uuml;ft ob die Antwort einen gesuchten String enth&auml;lt.
	 *
	 * @param antworten
	 *            Antwort des Faxgeraets
	 * @param antwort
	 *            Gesuchte Zeichenkette
	 * @param antwortwert
	 *            Wert der Antwort
	 * @param fehlermeldung
	 *            Fehlermeldung, wenn gesuchte Zeichenkette nicht gefunden ist
	 * @return true, wenn gesuchte Zeichenkette gefunden ist
	 * @throws CommunicationException
	 *             wird geworfen, wenn gesuchte Zeichenkette nicht gefunden ist
	 */
	public boolean ueberPruefeAntwort(final Collection<String> antworten, final String antwort,
			final String antwortwert, final String fehlermeldung) throws CommunicationException {
		for (final String s : antworten) {
			if (s.startsWith(antwort)) {
				String wert = s.substring(antwort.length()).trim();
				if (wert.startsWith("<SP>")) {
					wert = wert.substring(4);
				}

				if (wert.equalsIgnoreCase(antwortwert)) {
					return true;
				}
			}
		}
		throw new CommunicationException(fehlermeldung);
	}
}
